"""Forcing schemes (:mod:`fluidsim.base.forcing`)
=======================================================


Provides:

.. autosummary::
   :toctree:

   base
   specific
   anisotropic

"""

from .base import ForcingBase, ForcingBasePseudoSpectral
