"""Base classes for writing solvers
===================================


Provides:

.. autosummary::
   :toctree:

   solvers
   params
   setofvariables
   state
   init_fields
   time_stepping
   output
   forcing
   preprocess
   sphericalharmo

Third party solver interfaces:

.. autosummary::
   :toctree:

   basilisk
   dedalus

"""
