"""Mechanism to extend a Simul class with just a simple class
=============================================================

.. autofunction:: extend_simul_class

.. autoclass:: SimulExtender
   :members:
   :private-members:

.. autosummary::
   :toctree:

   spatial_means_regions_milestone

"""

from fluidsim_core.extend_simul import (
    extend_simul_class,
    SimulExtender,
    _extend_simul_class_from_path,
)
