"""Basilisk solvers (:mod:`fluidsim.base.basilisk`)
===================================================
A reference implementation to use 2D solvers in `Basilisk <http://basilisk.fr>`__ via
FluidSim.

Provides:

.. autosummary::
   :toctree:

    solver
    state
    operators
    time_stepping
    output

"""
