"""1-layer Shallow-Water solvers (:mod:`fluidsim.solvers.sw1l`)
=====================================================================

Provides

.. autosummary::
   :toctree:

   solver
   state
   operators
   forcing
   init_fields
   output

Modyfied solvers

.. autosummary::
   :toctree:

   exactlin
   modified
   onlywaves

"""
