"""Navier-Stokes 2D solvers (:mod:`fluidsim.solvers.ns2d`)
==========================================================

Base package for the Navier-Stokes 2D solvers. The main solver defined
in :mod:`fluidsim.solvers.ns2d.solver` is pseudo-spectral.

The modules for this ns2d solvers are:

.. autosummary::
   :toctree:

   solver
   state
   init_fields
   output
   forcing

There are also a packages for other Navier-Stokes 2D solvers:

.. autosummary::
   :toctree:

   strat
   bouss

"""
