"""Stratified Navier-Stokes 2D (:mod:`fluidsim.solvers.ns2d.strat`)
===================================================================

Provides:

.. autosummary::
   :toctree:

   solver
   state
   init_fields
   output
   forcing

"""
