"""Utilities for FluidSim
=========================

Provides:

.. autosummary::
   :toctree:

   util
   testing
   console
   mini_oper_modif_resol

"""

from .util import times_start_end_from_path
