"""Navier-Stokes 2D Boussinesq solvers (:mod:`fluidsim.solvers.ns2d.bouss`)
===========================================================================

Base package for the Navier-Stokes 2D under the Boussinesq approximation
solvers. The main solver defined in :mod:`fluidsim.solvers.ns2d.bouss.solver`
is pseudo-spectral.

The modules for this ns2d.bouss solvers are:

.. autosummary::
   :toctree:

   solver
   state

"""
