"""Dedalus solvers (:mod:`fluidsim.base.dedalus`)
=================================================

A reference implementation to use 2D solvers in `Dedalus
<https://dedalus-project.readthedocs.io>`__ via FluidSim.

Provides:

.. autosummary::
   :toctree:

    solver
    state
    operators
    time_stepping
    output

"""
