"""Variable container (:mod:`fluidsim.base.setofvariables`)
===========================================================

Provides:

.. autoclass:: SetOfVariables
   :members:
   :private-members:

"""

from fluiddyn.calcul.setofvariables import SetOfVariables
