"""NS2D over a sphere (:mod:`fluidsim.solvers.sphere.ns2d`)
===========================================================

Solving the incompressible Navier-Stokes equations over a sphere.

.. autosummary::
   :toctree:

   solver

"""
