"""Navier-Stokes 3D solvers (:mod:`fluidsim.solvers.ns3d`)
==========================================================

Submodules and packages
-----------------------

.. autosummary::
   :toctree:

   solver
   state
   init_fields
   forcing
   output

Derived solvers
---------------

.. autosummary::
   :toctree:

   strat
   bouss

"""
