"""Solving equations over a sphere (:mod:`fluidsim.solvers.sphere`)
===================================================================

Package containing the source code of the solvers to solve equations over a
sphere.

.. autosummary::
   :toctree:

   ns2d
   sw1l

"""
